#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/binfmts.h>
#include <linux/highmem.h>
#include <linux/file.h>
#include <linux/fs_struct.h>
#include <linux/namei.h>
#include <linux/path.h>
#include <linux/list.h>
#include <linux/version.h>
#include <linux/xattr.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)
#include <uapi/linux/mman.h>
#else
#include <uapi/asm-generic/mman-common.h>
#endif

#include <linux/uos_elfverify.h>


#define MODULENAME "elfverify_extend"
#define MODULEVERSION "v0.0.1"


int java_check_file_security_attr(char* cwd_path, char* token, char **abs_path){
	char xattr[8] = {0};
	struct file* file = NULL;
	static int intxattr = 0;
	int ret = 0;

	memset(*abs_path, 0, PATH_MAX + 1);
	if (token[0] != '/'){
		strcat(*abs_path, cwd_path);
		strcat(*abs_path, "/");
		strcat(*abs_path, token);
	}else{
		strcpy(*abs_path, token);
	}
	file = filp_open(*abs_path, O_RDONLY, 0);
	if (IS_ERR(file)) {
		// printk("[%s] can't get file abs path, need check again. path: %s\n", __func__, *abs_path);
		memset(*abs_path, 0, PATH_MAX + 1);
		return 0;
	}

	ret = vfs_getxattr(file->f_path.dentry, "security.integrity", xattr, 8);
	if (ret <= 0){
		goto out;
	}
	if (kstrtoint(xattr, 10, &intxattr) != 0){
		goto out;
	}
	//todo: 由于只能返回一个文件，加入策略验证过滤更准确
	if (intxattr == 0x1){ //属于系统应用且可信，放行
		memset(*abs_path, 0, PATH_MAX + 1);
	}
out:
	filp_close(file, 0);
	return 0;
}


int get_java_abs_file_path(char* script_file, char** abs_file)
{
	char* cwd_path = NULL;
	char* path_buf = NULL;
	char* token = NULL;

	path_buf = kzalloc(PATH_MAX, GFP_KERNEL);
	if (unlikely(path_buf == NULL)) {
		return -ENOMEM;
	}
	cwd_path = d_path(&(current->fs->pwd), path_buf, PATH_MAX);
	// printk("cwd_path = %s\n", cwd_path);
	
	//由于script_filter只支持返回单个文件，因此提前判断各个jar包文件标签，返回第一个不可信文件
	while ((token = strsep(&script_file,":")) != NULL){
		java_check_file_security_attr(cwd_path, token, abs_file);
		if (strlen(*abs_file) != 0){
			goto out;
		}
	}

out: 
	if (path_buf) {
		kfree(path_buf);
		path_buf = NULL;
	}
	return 0;
}


/*
 * First determine whether the execution is in the current path, and
 * then determine whether it is in the environment variable.
 */
int get_abs_file_path(const char *env_path, char* script_file,int find_env, char** abs_file)
{
	char* full_path = NULL;
	char* cwd_path = NULL;
	char* path_buf = NULL;
	struct file* file = NULL;
	char* env = NULL;
	char* tmp_str = NULL;

	full_path = *abs_file;
	memset(full_path, 0, PATH_MAX+1);

	path_buf = kzalloc(PATH_MAX + 1, GFP_KERNEL);
	if (unlikely(path_buf == NULL)) {
		return -ENOMEM;
	}

	if (script_file[0] == '/'){
		strcpy(full_path, script_file);
		file = filp_open(full_path, O_RDONLY, 0);
	}else{ // need get abs path
		cwd_path = d_path(&(current->fs->pwd), path_buf, PATH_MAX);
		// printk("cwd_path = %s , exec env = %s\n", cwd_path, env_path);
		//获取文件绝对路径
		strcat(full_path, cwd_path);
		strcat(full_path, "/");
		strcat(full_path, script_file);
		file = filp_open(full_path, O_RDONLY, 0);
		if (IS_ERR(file) && find_env ) {
			tmp_str = (char *)env_path;
			// PATH=/usr/local/bin:/usr/bin:/bin:/usr/local/games:/usr/games:/sbin:/usr/sbin
			while ((env = strsep(&tmp_str, ":")) != NULL){
				if (env[0] != '/'){
					continue;
				}
				memset(full_path, 0, PATH_MAX + 1);
				strcat(full_path, env);
				strcat(full_path, "/");
				strcat(full_path, script_file);
				printk("get env, full_path = %s\n",full_path);
				file = filp_open(full_path, O_RDONLY, 0);
				if (!IS_ERR(file)){
					break;
				}
			}
		}
	}
	if (IS_ERR(file)) {
		memset(full_path, 0, PATH_MAX + 1);
		printk("[%s] can't get file abs path. path: %s\n", __func__, script_file);
	}else{
		filp_close(file, 0);
	}
	
	if (path_buf) {
		kfree(path_buf);
		path_buf = NULL;
	}
	return 0;
}

int parse_bash_param(const int argc, char *argv[], const char *env_path, char **outfile)
{
	int arg_index = 0;
	char *arg_str = NULL;
	char *script_file = NULL;
	char* abs_file = NULL;

	// pr_info(" into parse_bash_param\n");
	/*
		找到脚本文件 
		如果是 "-c"  直接返回;
		如果arg 是 "-/+ " ,find_next+1 。另外如果arg+1 是“o/O” ,再次find_next+1; 
		如果arg+1 是"-" （长参数），find_next+1;
	*/
	for(arg_index = 1; arg_index < argc; ){
		// printk("argv[%d] = %s\n",arg_index , argv[arg_index]);
		arg_str = argv[arg_index];
		if(arg_str[0] == '-' || arg_str[0] == '+'){
			switch(arg_str[1])
			{
			case 'c': //skip
				memset(*outfile, 0, PATH_MAX + 1);
				return 0;
			case 'o':
			case 'O':
				arg_index += 2;
				break;
			default:
				arg_index++;
			}
		}else{
			script_file = arg_str;
			// printk("bash get file name form argv: %s\n", script_file);
			break;
		}
	}
	if (script_file == NULL){
	    return 0;
	}
	abs_file = kzalloc(PATH_MAX + 1, GFP_KERNEL);
	if (unlikely(abs_file == NULL)) {
		return -ENOMEM;
	}
	get_abs_file_path(env_path, script_file, 1, &abs_file);
	// printk("abs_file = %s\n", abs_file);
	if (strlen(abs_file) != 0){
		strcpy(*outfile, abs_file);
	}
	kfree(abs_file);
	return 0;
}

int parse_perl_param(const int argc, char *argv[], const char *env_path, char **outfile)
{
	int arg_index = 0;
	char *arg_str = NULL;
	char *script_file = NULL;
	char* abs_file = NULL;
	int find_env_flag =0;

	// pr_info(" into parse_perl_param\n");

	for(arg_index = 1; arg_index < argc; ){
		// printk("argv[%d] = %s\n",arg_index , argv[arg_index]);
		arg_str = argv[arg_index];
		if(arg_str[0] == '-'){
			switch(arg_str[1])
			{
			case 'E': //skip
			case 'e': //skip
				memset(*outfile, 0, PATH_MAX + 1);
				return 0;
			case 'S':
				find_env_flag = 1;
			default:
				arg_index++;
			}
		}else{
			script_file = arg_str;
			// printk("perl get file name form argv: %s\n", script_file);
			break;
		}
	}
	if (script_file == NULL){
	    return 0;
	}
	abs_file = kzalloc(PATH_MAX + 1, GFP_KERNEL);
	if (unlikely(abs_file == NULL)) {
		return -ENOMEM;
	}
	get_abs_file_path(env_path, script_file, find_env_flag, &abs_file);
	// printk("abs_file = %s\n", abs_file);
	if (strlen(abs_file) != 0){
		strcpy(*outfile, abs_file);
	}
	kfree(abs_file);
	return 0;
}

int parse_python_param(const int argc, char *argv[], const char *env_path, char **outfile)
{
	int arg_index = 0;
	char *arg_str = NULL;
	char *script_file = NULL;
	char* abs_file = NULL;

	// pr_info(" into parse_python_param, argc=%d\n",argc);
	/*
		找到脚本文件 
		如果是 "-c"  后面是命令，没有文件，需要将此事件上报
		如果arg 是 "-mWX" ,后面参数可有可无空格，要根据长度判断跳过1或2个参数。 
	*/
	for(arg_index = 1; arg_index < argc; ){
		// printk("argv[%d] = %s\n",arg_index , argv[arg_index]);
		arg_str = argv[arg_index];
		if(arg_str[0] == '-'){
			switch(arg_str[1])
			{
			case 'c': //zhouzilong(todo): 暂时不管，之后直接强制上报
				memset(*outfile, 0, PATH_MAX + 1);
				return 0;
			case 'm': //zhouzilong(todo): python -m mod, 可以直接运行模块，需要查找模块所在文件。
			case 'W':
			case 'X':
				if(strlen(arg_str)>2){
				    arg_index += 1;
				}else{
				    arg_index += 2;
				}
				break;
			case '-':
				if(!strcmp(arg_str, "--check-hash-based-pycs")){
				    arg_index += 2;
				}else{
				    arg_index++;
				}
				break;
			default:
				arg_index++;
			}
		}else{
			script_file = arg_str;
			// printk("python get file name form argv: %s\n", script_file);
			break;
		}
	}
	if (script_file == NULL){
	    return 0;
	}
	abs_file = kzalloc(PATH_MAX + 1, GFP_KERNEL);
	if (unlikely(abs_file == NULL)) {
		return -ENOMEM;
	}
	get_abs_file_path(env_path, script_file, 0, &abs_file);
	// printk("abs_file = %s\n", abs_file);
	if (strlen(abs_file) != 0){
		strcpy(*outfile, abs_file);
	}
	kfree(abs_file);
	return 0;
}

int parse_java_param(const int argc, char *argv[], const char *env_path, char **outfile)
{
	int arg_index = 0;
	char *arg_str = NULL;
	char *script_file = NULL;
	char* abs_file = NULL;

	// pr_info(" into parse_java_param\n");

	for(arg_index = 1; arg_index < argc; arg_index++){
		// printk("argv[%d] = %s\n",arg_index , argv[arg_index]);
		arg_str = argv[arg_index];
		if(!strcmp(arg_str, "-classpath")){
		    script_file = argv[arg_index+1];//解析多文件
			break;
		}else if(!strcmp(arg_str, "-cp")){
		    script_file = argv[arg_index+1];
			break;
		}
		if(!strcmp(arg_str, "--class-path")){
		    script_file = argv[arg_index+1];
			break;
		}
		if(!strcmp(arg_str, "-jar")){
			script_file = argv[arg_index+1];
			break;
		}
	}
	if (script_file == NULL){
	    return 0;
	}

	abs_file = kzalloc(PATH_MAX + 1, GFP_KERNEL);
	if (unlikely(abs_file == NULL)) {
		return -ENOMEM;
	}
	memset(abs_file, 0, PATH_MAX+1);

	//处理多文件 , 遍历打开文件和读扩展属性，返回需要验证的文件。
	get_java_abs_file_path(script_file, &abs_file);
	// printk("java  abs_file need verify = %s\n", abs_file);
	if (strlen(abs_file) != 0){
		strcpy(*outfile, abs_file);
	}
	kfree(abs_file);
	return 0;
}


struct script_options{
	char name[256];
	int (*parse_param)(const int argc, char *argv[], const char *env_path, char **outfile);
}script_ops[] = {
	{"/usr/bin/bash", parse_bash_param},
	{"/usr/bin/dash", parse_bash_param},
	{"/usr/bin/perl", parse_perl_param},
	{"/usr/bin/python3.7", parse_python_param},
	{"/usr/bin/python3.7m", parse_python_param},
	{"/usr/bin/python2.7", parse_python_param},
	{"/usr/lib/jvm/java-11-openjdk-amd64/bin/java", parse_java_param},
	{"/usr/lib/jvm/java-11-openjdk-arm64/bin/java", parse_java_param},
	{"/usr/lib/jvm/java-11-openjdk-loongarch64/bin/java", parse_java_param},
};
#define SCRIPT_OPS_SIZE sizeof(script_ops) / sizeof(script_ops[0])


int dispatch_inter_type(const char *script_type, const int argc, char **argv, const char *env_path, char **outfile){
	int i;

	// pr_info(" into dispatch_inter_type\n");
	
	for (i = 0; i< SCRIPT_OPS_SIZE; i++){
		if(!strcmp(script_type, script_ops[i].name)){
			script_ops[i].parse_param(argc, argv, env_path, outfile);
			break;
		}
	}
	return 0;
}

static void exec_guard_put_page(struct page *page)
{
#ifdef CONFIG_MMU
	put_page(page);
#endif
}

static struct page *exec_guard_get_page(const struct linux_binprm *bprm, unsigned long pos)
{
	struct page *page;

	#ifdef CONFIG_MMU
		if (get_user_pages_remote(current, bprm->mm, pos, 1,
			FOLL_FORCE, &page, NULL, NULL) <= 0)
			return NULL;
	#else
		page = bprm->page[pos / PAGE_SIZE];
	#endif

	return page;
}

int exec_guard_dump_page(const char *script_type, const struct linux_binprm *bprm, char **outfile)
{
	int ret;
	int argc, envc;
	unsigned long pos, offset;
	struct page *page;
	char *kaddr;
	char *param_buffer = NULL;
	char *param_buffer_tmp = NULL;
	int i, count;
	char **argv = NULL;
	char *script_exec_env = NULL;

	argc = bprm->argc;
	envc = bprm->envc;
	pos = bprm->p;
	i = 0;
	count = 0;

	// pr_info("into filter_params filename=%s\n",bprm->filename);
	
	if(argc == 1) //没有任何参数
		goto out;

	argv = (char **)kzalloc(sizeof(char *)*argc, GFP_KERNEL);
	if (unlikely(argv == NULL))
		return -ENOMEM;

	param_buffer = kzalloc(PAGE_SIZE + 1, GFP_KERNEL);
	if (unlikely(param_buffer == NULL))
		goto out;
			
	do {
		page = exec_guard_get_page(bprm, pos);//获取page
		if(!page){
			ret = 0;
			goto out;
		}
		kaddr = kmap_atomic(page);//获取页映射地址
		offset = pos & ~PAGE_MASK; //得到page内的偏移（起始地址）	
		// pr_info(" get offset sum= %d\n", argc + envc);
		for( ;offset < PAGE_SIZE && count < argc + envc; offset++, pos++){//遍历页地址找到需要的argv和env
			
			if (i != 0 && i % PAGE_SIZE == 0){ //若不够，动态申请存放参数的空间
				int argvLen;
				// pr_info(" reload page_size, strlen(param_buffer)= %ld\n",strlen(param_buffer));
				argvLen = strlen(param_buffer) + PAGE_SIZE + 1;
				param_buffer_tmp = kzalloc(argvLen, GFP_KERNEL);
				if (unlikely(param_buffer_tmp == NULL))
					goto out;
				memcpy(param_buffer_tmp, param_buffer, strlen(param_buffer));
				kfree(param_buffer);
				param_buffer = param_buffer_tmp;
			}

			param_buffer[i] = kaddr[offset];
			// pr_info(" kaddr[%d] = %c; param_buffer[%d] = %c\n", offset, kaddr[offset], i, param_buffer[i]);
			i++;
			if (kaddr[offset] == '\0'){ //从page中读到了一个完整字符串; page前存argv,后存env
				if (count < argc){
					argv[count] = kzalloc(strlen(param_buffer) + 1, GFP_KERNEL); //申请空间存放参数
					if (unlikely(argv[count] == NULL))
						goto out;
					memcpy(argv[count], param_buffer, strlen(param_buffer));
				}else if (strncmp(param_buffer, "PATH=", 5) == 0){ //读取env PATH值
					// pr_info("(PATH) param_buffe[%d] = %s\n", count, param_buffer);
					script_exec_env = kzalloc(strlen(param_buffer) + 1, GFP_KERNEL);
    				if (unlikely(script_exec_env == NULL))
    					goto out;
	    			strcpy(script_exec_env , param_buffer + 5);
	    			break;
				}
				// pr_info(" strlen(param_buffer)= %ld, param_buffe[%d] = %s\n", strlen(param_buffer),count, param_buffer);
				i=0; //clear
				memset(param_buffer, 0, strlen(param_buffer));
				count ++;
			}
		}
	kunmap_atomic(kaddr);
	exec_guard_put_page(page);
	}while (offset == PAGE_SIZE);

	dispatch_inter_type(script_type, argc, argv, script_exec_env, outfile);
	if (strlen(*outfile) != 0){
		pr_info("[elfverify_extend_module] filter_params file=%s\n", *outfile);
	}

out:
	if (param_buffer) {
    	kfree(param_buffer);
    	param_buffer = NULL;
    }
    if (script_exec_env) {
    	kfree(script_exec_env);
    	script_exec_env = NULL;
    }
    if (argv) {
		int y=0;
		for(y=0; y<argc; y++){
			if(argv[y])
				kfree(argv[y]);
		}
    	kfree(argv);
    }

	return 0;
}

static struct uos_elfverify_operations entry={
	.script_filter = exec_guard_dump_page,
	.sign_verify = NULL,
};


int __init init_module(void)
{
	int i;
	struct file *file = NULL;
	pr_info("%s(%s) start\n", MODULENAME, MODULEVERSION);
	for(i = 0; i < SCRIPT_OPS_SIZE; i++){
		file = filp_open(script_ops[i].name, O_RDONLY, 0);
		if (!IS_ERR(file)) {
			uos_script_guard_register(script_ops[i].name);
			filp_close(file, NULL);
			file = NULL;
		}
	}
	uos_elfverify_backend_register(&entry);
	return 0;
}

void __exit cleanup_module(void)
{
	pr_info("%s(%s) exit\n", MODULENAME, MODULEVERSION);
	uos_elfverify_backend_unregister();
	uos_script_guard_unregister();

	
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("zhozilong");
MODULE_DESCRIPTION("This module provides extended functionality for elfverify");
